import axios from "axios";
import { APIException, NetworkConnectionException } from "../Exception";
import { parseWithDate } from "./json-util";
const networkInterceptor = {};
export function setRequestHeaderInterceptor(_) {
    networkInterceptor.request = _;
}
export function setResponseHeaderInterceptor(_) {
    networkInterceptor.response = _;
}
axios.defaults.transformResponse = (data, headers) => {
    const contentType = headers["content-type"];
    if (contentType && contentType.startsWith("application/json")) {
        return parseWithDate(data);
    }
    return data;
};
axios.interceptors.response.use(response => response, e => {
    // eslint-disable-next-line no-prototype-builtins
    if (e && typeof e === "object" && e.hasOwnProperty("isAxiosError")) {
        const error = e;
        const requestURL = error.config.url || "-";
        if (error.response) {
            // Try to get server error message/ID/code from response
            const responseData = error.response.data;
            const errorId = responseData && responseData.id ? responseData.id : null;
            const errorCode = responseData && responseData.errorCode ? responseData.errorCode : null;
            if (!errorId && (error.response.status === 502 || error.response.status === 504)) {
                // Treat "cloud" error as Network Exception, e.g: gateway issue, load balancer unconnected to application server
                // Note: Status 503 is maintenance
                throw new NetworkConnectionException(`gateway error (${error.response.status})`, requestURL, error.message);
            }
            else {
                const errorMessage = responseData && responseData.message ? responseData.message : `[No response message]`;
                throw new APIException(errorMessage, error.response.status, requestURL, responseData, errorId, errorCode);
            }
        }
        else {
            /**
             * It could be network failure, or CORS pre-flight failure. We cannot distinguish here.
             * Ref: https://github.com/axios/axios/issues/838
             */
            throw new NetworkConnectionException(`failed to connect to ${requestURL}`, requestURL, error.message);
        }
    }
    else {
        throw new NetworkConnectionException(`unknown network error`, `[No URL retrieved]`, e.toString());
    }
});
export async function ajax(method, path, pathParams, request) {
    const fullURL = urlParams(path, pathParams);
    const config = { method, url: fullURL };
    if (method === "GET" || method === "DELETE") {
        config.params = request;
    }
    else if (method === "POST" || method === "PUT" || method === "PATCH") {
        config.data = request;
    }
    const requestHeaderMap = {
        "Content-Type": "application/json",
        Accept: "application/json",
    };
    if (networkInterceptor.request) {
        await networkInterceptor.request(requestHeaderMap);
    }
    config.headers = requestHeaderMap;
    const response = await axios.request(config);
    const responseHeaderMap = response.headers;
    if (networkInterceptor.response) {
        await networkInterceptor.response(responseHeaderMap);
    }
    return response.data;
}
export function uri(path, request) {
    const config = { method: "GET", url: path, params: request };
    return axios.getUri(config);
}
export function urlParams(pattern, params) {
    if (!params) {
        return pattern;
    }
    let url = pattern;
    Object.entries(params).forEach(([name, value]) => {
        const encodedValue = encodeURIComponent(value.toString());
        url = url.replace(":" + name, encodedValue);
    });
    return url;
}
//# sourceMappingURL=network.js.map